import json

import osmium


golf_tags = {
    ("golf", "hole"),
    ("golf", "tee"),
    ("golf", "green"),
    ("golf", "fairway"),
    ("golf", "rough"),
    ("golf", "bunker"),
    ("golf", "water_hazard"),
    ("golf", "lateral_water_hazard"),
    ("natural", "wood"),
}


class GolfHandler(osmium.SimpleHandler):
    """
    Handler to parse an OpenStreetMap XML file and construct a GeoJSON-like feature
    collection (dictionary) of all its golf-related objects.

    Usage
    -----
    >>> handler = GolfHandler()
    >>> handler.apply_file(map_path, locations=True)
    >>> json.dumps(handler.feature_collection)
    """

    def __init__(self):
        super().__init__()
        self.jsonfab = osmium.geom.GeoJSONFactory()
        self.feature_collection = {"type": "FeatureCollection", "features": []}

    def way(self, w):
        if w.is_closed():
            return
        self.add_feature(w, "linestring")

    def area(self, a):
        self.add_feature(a, "multipolygon")

    def add_feature(self, obj, geom_type):
        match geom_type:
            case "linestring":
                factory_method = self.jsonfab.create_linestring
            case "multipolygon":
                factory_method = self.jsonfab.create_multipolygon
            case _:
                raise ValueError(f"{geom_type=} not supported")
        geom_string = factory_method(obj)
        for key, value in golf_tags:
            if obj.tags.get(key) == value:
                geom_string = factory_method(obj)
                feature = {
                    "type": "Feature",
                    "geometry": json.loads(geom_string),
                    "properties": dict(obj.tags),
                }
                self.feature_collection["features"].append(feature)



from dataclasses import dataclass, field

import geopandas as gpd
import shapely

from pitchmark.plotting import chart_course


@dataclass
class Hole:
    """A representation of a golf hole."""

    hole_number: int
    name: str = ""
    path: shapely.LineString | None = None
    gdf: gpd.GeoDataFrame | None = field(default=None, repr=False)
    mesh: gpd.GeoDataFrame | None = field(default=None, repr=False)

    def chart(self, **kwargs):
        return (
            chart_course(self.gdf, **kwargs)
            .properties(title=f"{self.hole_number} - {self.name}")
            .configure_legend(disable=True)
        )



"""A subpackage for parsing LiDAR files."""

# read version from installed package
from importlib.metadata import version

__version__ = version("pitchmark")

# Populate namespace
from pitchmark.lidar import las



from dataclasses import dataclass, field
import json
import logging
import operator
import warnings

import geopandas as gpd
import laspy
import pandas as pd
import pyproj
import shapely

from pitchmark.geom import (
    delaunay3d,
    gdf_from_mesh,
    polygon_from_geojson,
    prepared_shell,
    simplified_mesh,
)
from pitchmark.hole import Hole
from pitchmark.plotting import chart_course


@dataclass
class Course:
    """A representation of a golf course."""

    holes: list[Hole] = field(default_factory=list)
    greens: list[shapely.Polygon] = field(default_factory=list)
    tees: list[shapely.Polygon] = field(default_factory=list)
    fairways: list[shapely.Polygon] = field(default_factory=list)
    bunkers: list[shapely.Polygon] = field(default_factory=list)
    rough: list[shapely.Polygon] = field(default_factory=list)
    water: list[shapely.Polygon] = field(default_factory=list)
    woods: list[shapely.Polygon] = field(default_factory=list)
    proj_string: str | None = field(init=False)
    transformer_to_local: pyproj.Transformer | None = field(init=False, repr=False)
    gdf: gpd.GeoDataFrame = field(init=False, repr=False)

    def __post_init__(self):
        self.holes.sort(key=operator.attrgetter("hole_number"))
        # transverse Mercator centered on 1st tee with WGS84 datum, in yards
        try:
            lon_0, lat_0 = self.holes[0].path.coords[0]
            self.proj_string = f"+proj=tmerc +{lon_0=} +{lat_0=} +ellps=WGS84 +units=yd"
            self.transformer_to_local = pyproj.Transformer.from_crs(
                "EPSG:4326", self.proj_string, always_xy=True
            )
        except IndexError:
            warnings.warn(
                "No first tee present, could not construct projection string.",
                stacklevel=2,
            )
            self.proj_string = None
        data = {
            "name": pd.Categorical(
                [
                    "water_hazard",
                    "bunkers",
                    "green",
                    "tee",
                    "fairway",
                    "woods",
                    "rough",
                ],
            ),
            "course_area": pd.Categorical(
                [
                    "penalty_area",
                    "bunker",
                    "putting_green",
                    "teeing_area",
                    "general_area",
                    "general_area",
                    "general_area",
                ],
                categories=[
                    "penalty_area",
                    "bunker",
                    "putting_green",
                    "teeing_area",
                    "general_area",
                ],
                ordered=True,
            ),
            "ground_cover": pd.Categorical(
                [
                    "water",
                    "sand",
                    "green",
                    "short_grass",
                    "short_grass",
                    "woods",
                    "long_grass",
                ],
                categories=[
                    "water",
                    "sand",
                    "green",
                    "short_grass",
                    "woods",
                    "long_grass",
                ],
                ordered=True,
            ),
        }
        course_gdf = gpd.GeoDataFrame(
            data,
            geometry=[
                shapely.MultiPolygon(x)
                for x in [
                    self.water,
                    self.bunkers,
                    self.greens,
                    self.tees,
                    self.fairways,
                    self.woods,
                    self.rough,
                ]
            ],
            crs="EPSG:4326",  # WGS84 in longitude/latitude (degrees)
        ).explode(ignore_index=True)
        if self.proj_string is not None:
            self.gdf = course_gdf.to_crs(self.proj_string)
            self.populate_hole_gdfs()
            self.trim_to_features_in_play()
        else:
            self.gdf = course_gdf

    @classmethod
    def from_featurecollection(cls, fc):
        course_dict = dict(fc)
        holes, greens, tees, fairways, bunkers, rough, water, woods = [
            [] for _ in range(8)
        ]
        for feature in course_dict["features"]:
            geometry = feature["geometry"]
            properties = feature["properties"]
            geom_string = json.dumps(geometry)
            if properties.get("golf") == "hole":
                holes.append(
                    Hole(
                        int(properties.get("ref")),
                        properties.get("name"),
                        shapely.from_geojson(geom_string),
                    )
                )
            if properties.get("golf") == "green":
                greens.append(polygon_from_geojson(geom_string))
            if properties.get("golf") == "tee":
                tees.append(polygon_from_geojson(geom_string))
            if properties.get("golf") == "fairway":
                fairways.append(polygon_from_geojson(geom_string))
            if properties.get("golf") == "bunker":
                bunkers.append(polygon_from_geojson(geom_string))
            if properties.get("golf") == "rough":
                rough.append(polygon_from_geojson(geom_string))
            if properties.get("golf") in ("water_hazard", "lateral_water_hazard"):
                water.append(polygon_from_geojson(geom_string))
            if properties.get("natural") == "wood":
                woods.append(polygon_from_geojson(geom_string))

        return cls(holes, greens, tees, fairways, bunkers, rough, water, woods)

    def chart(self, **kwargs):
        return chart_course(self.gdf, **kwargs).configure_legend(disable=True)

    def mask_hole(
        self,
        hole_path,
        *,
        tee_buffer=10.0,
        fairway_buffer=20.0,
        green_buffer=30.0,
        approach_buffer=50.0,
    ):
        fairways = self.gdf[
            (self.gdf.name == "fairway")
            & (self.gdf["geometry"].apply(shapely.intersects, args=(hole_path,)))
        ]
        green = self.gdf[
            (self.gdf.name == "green")
            & (self.gdf["geometry"].apply(shapely.intersects, args=(hole_path,)))
        ]
        shot_locations = [shapely.Point(hole_path.coords[0]).buffer(tee_buffer)] + [
            shapely.buffer(shapely.Point(x), approach_buffer)
            for x in hole_path.coords[1:]
        ]
        shot_connectors = []
        for x, y in zip(shot_locations[:-1], shot_locations[1:]):
            mp = shapely.MultiPolygon([x, y])
            shot_connectors.append(mp.convex_hull)
        shots_envelope = shapely.unary_union(shot_connectors)
        mask = shapely.unary_union(
            (
                fairways.unary_union.buffer(fairway_buffer),
                green.unary_union.buffer(green_buffer),
                shots_envelope,
            )
        )
        return mask

    def populate_hole_gdfs(self, **kwargs):
        for hole in self.holes:
            hole_path = shapely.ops.transform(
                self.transformer_to_local.transform, hole.path
            )
            mask = self.mask_hole(hole_path, **kwargs)
            hole.gdf = self.gdf.clip(mask)

    def trim_to_features_in_play(self):
        indices_in_play = pd.concat(
            [hole.gdf for hole in self.holes],
        ).index.unique()
        in_play = self.gdf.iloc[indices_in_play]
        self.gdf = in_play

    def populate_hole_meshes(
        self, ground_path, *, buffer_distance=1.0, merge_close=0.25, smooth_iters=10
    ):
        """
        Read a LAS/LAZ file with an elevation point cloud, construct 3D triangle meshes,
        smoothen the mesh, and add mesh info to each hole.

        Parameters:
        -----------
        ground_path: pathlib.Path
            Location of the point cloud data file. Must contain ground level data and
            is assumed to extend to include the whole course. It is recommended to pre-
            process raw files to extract LAS data classified as "ground" and if required
            merge multiple files into one before reading this.
        buffer_distance: float, default 1.0
            Extent by which to enlarge the area defined by the hole geometry. Points in
            the point cloud are included if they are inside the buffer of the hole.
        merge_close: float, default 0.25
            Neighbouring points closer together than this distance are merged together.
        smooth_iters: int, default 10
            Number of times to perform the smoothening algorithm.

        See also:
        ---------
        lidar.las.clip_to_geoseries
        """
        ground = laspy.read(ground_path)
        logging.info(f"Populating hole meshes from {ground_path}:")
        ground_crs = ground.header.parse_crs()
        local_crs = self.gdf.crs
        transformer = pyproj.Transformer.from_crs(ground_crs, local_crs)

        for hole in self.holes:
            logging.info(f"Preparing hole {hole.hole_number}...")
            masked_area_cloud = prepared_shell(
                hole.gdf.geometry,
                distance=buffer_distance,
                crs_to=ground_crs,
            )
            hole_ground = ground[shapely.contains_xy(masked_area_cloud, ground.xyz)]
            xyz = hole_ground.xyz

            masked_area_local = prepared_shell(
                hole.gdf.geometry,
                distance=buffer_distance,
            )
            logging.info("Constructing Delaunay triangles...")
            hole_triangles = delaunay3d(
                xyz, transformer=transformer, shell=masked_area_local
            )
            logging.info("Constructing 3D mesh...")
            mesh = simplified_mesh(
                hole_triangles, merge_close=merge_close, smooth_iters=smooth_iters
            )
            logging.info("Aggregating dataframe...")
            hole.mesh = gdf_from_mesh(mesh, crs=local_crs)
        logging.info("Done!")



"""A package for exploring golf shot strategy."""

# read version from installed package
from importlib.metadata import version

__version__ = version("pitchmark")


# populate package namespace
from pitchmark.course import Course
from pitchmark.hole import Hole
from pitchmark.plotting import chart_course



import geopandas as gpd
import numpy as np
import open3d as o3d
import pyproj
import shapely


def polygon_from_geojson(geom_string):
    multipolygon = shapely.from_geojson(geom_string)
    geoms = multipolygon.geoms
    if len(geoms) != 1:
        raise ValueError(
            "MultiPolygon must contain exactly one polygon for lossless unpacking"
        )
    polygon = geoms[0]
    return polygon


def prepared_shell(geoseries, *, distance=0.0, crs_to=None):
    """
    Construct a shell (buffer) around the (unary union of the) geometries in a given
    Geoseries, optionally transform it to a different CRS, and prepare it to improve
    computational efficiency of subsequent operations.
    """
    masked_area = geoseries.unary_union.buffer(distance)

    if crs_to is not None:
        transformer = pyproj.Transformer.from_crs(geoseries.crs, crs_to, always_xy=True)
        masked_area = shapely.ops.transform(transformer.transform, masked_area)

    shapely.prepare(masked_area)
    return masked_area


def delaunay3d(xyz, *, transformer=None, shell=None):
    """
    Construct 2.5D Delaunay triangulation from an X, Y, Z array of 3D point cloud data,
    and optionally project to a different CRS and/or remove triangles outside a
    predefined shell.

    Parameters:
    xyz: np.array
        Point cloud data.
    transformer: pyproj.Transformer or None, default None
    shell: shapely.Polygon or None, default None
        For best results, the shell should be prepared before calling this function.
    """
    if transformer is not None:
        xyz = np.array(list(transformer.itransform(xyz)))
    points = shapely.MultiPoint(xyz)
    dt = shapely.delaunay_triangles(points)
    if shell is None:
        return dt
    triangles = [t for t in dt.geoms if t.within(shell)]
    return triangles


def simplify_close_vertices(mesh, distance=0.001):
    """
    Simplify an Open3D TriangleMesh by merging close vertices and removing residual
    geometries.
    """
    mesh.merge_close_vertices(distance)
    mesh.remove_degenerate_triangles()
    mesh.remove_unreferenced_vertices()


def simplified_mesh(triangles, *, merge_close=0.25, smooth_iters=10):
    """
    Construct an Open3D mesh defined by a sequence of 2.5D shapely triangles, merge
    close vertices and smoothen out the surface.
    """
    vertices = o3d.utility.Vector3dVector(
        np.array([triangle.exterior.coords[:-1] for triangle in triangles]).reshape(
            (-1, 3)
        )
    )
    triangle_indices = o3d.utility.Vector3iVector(
        np.arange(len(vertices)).reshape((-1, 3))
    )
    mesh = o3d.geometry.TriangleMesh(vertices, triangle_indices)
    simplify_close_vertices(mesh, merge_close)
    for _ in range(smooth_iters):
        mesh = mesh.filter_smooth_taubin()
        simplify_close_vertices(mesh, merge_close)
    return mesh


def gdf_from_mesh(mesh, crs=None):
    """
    Construct a GeoDataFrame from an Open3D TriangleMesh.
    """
    triangles = shapely.polygons(np.asarray(mesh.vertices)[np.asarray(mesh.triangles)])
    mesh.compute_triangle_normals()
    normals = np.asarray(mesh.triangle_normals)
    centroids = np.array(
        [sum(np.array(triangle.exterior.coords[:-1])) / 3 for triangle in triangles]
    )
    gdf = gpd.GeoDataFrame(geometry=triangles, crs=crs)
    gdf[["x", "y", "z"]] = centroids
    gdf[["normal_x", "normal_y", "normal_z"]] = normals
    gdf["slope_heading"] = np.degrees(np.arctan2(gdf["normal_x"], gdf["normal_y"]))
    gdf["slope_heading"] = np.where(
        gdf["slope_heading"] < 0, gdf["slope_heading"] + 360, gdf["slope_heading"]
    )
    gdf["slope_grade"] = 100.0 * np.sqrt(1.0 - gdf["normal_z"] ** 2)
    return gdf



import altair as alt


_palette = {
    "water": "RoyalBlue",
    "sand": "Khaki",
    "green": "Aquamarine",
    "short_grass": "LawnGreen",
    "woods": "ForestGreen",
    "long_grass": "LimeGreen",
}


def chart_course(geodataframe, *, mode="ground_cover", tooltip=True):
    match mode:
        case "ground_cover":
            _domain = list(_palette.keys())
            _range = list(_palette.values())
        case _:
            raise ValueError(f"{mode=} not implemented")

    if tooltip is True:
        tooltip = ["name", "ground_cover", "course_area"]
    elif tooltip in (False, None):
        tooltip = alt.value(None)

    course_chart = (
        alt.Chart(geodataframe.sort_values(by=mode, ascending=False))
        .mark_geoshape()
        .encode(
            color=alt.Color(
                mode,
                scale=alt.Scale(
                    domain=_domain,
                    range=_range,
                ),
            ),
            tooltip=tooltip,
        )
        .project(type="identity", reflectY=True)
    )
    return course_chart



"""A subpackage for parsing OpenStreetMap files."""

# read version from installed package
from importlib.metadata import version

__version__ = version("pitchmark")

# Populate namespace
from pitchmark.osm.osm import golf_tags, GolfHandler



import enum

import laspy
import numpy as np
import shapely

from pitchmark.geom import prepared_shell


class LasClassification(enum.IntEnum):
    created_never_classified = 0
    unclassified = 1
    ground = 2
    low_vegetation = 3
    medium_vegetation = 4
    high_vegetation = 5
    building = 6
    low_point = 7
    noise = 7
    model_keypoint = 8
    mass_point = 8
    water = 9


def clip_to_geoseries(
    from_files,
    geoseries,
    *,
    to_file="out.laz",
    classification_filter=None,
    shell_distance=2.0
):
    """
    Read in multiple LAS/LAZ files, filter points with specified classifications,
    clip them to the extent of a specified GeoSeries with an additional buffer,
    and export to a single new file.

    The header is adapted from the first input file. Caution: this assumes that the
    input files are compatible with each other (i.e. using any of their headers should
    give the same result).

    Default classifications: Unclassified and Ground.
    """
    if classification_filter is None:
        classification_filter = [
            LasClassification.unclassified,
            LasClassification.ground,
        ]
    with laspy.open(from_files[0]) as handle:
        export_header = handle.header
    masked_area = prepared_shell(
        geoseries, distance=shell_distance, crs_to=export_header.parse_crs()
    )

    # file by file, extract raw LAS points data (stored in ndarrays), and prepare
    # a list of these arrays that is concatenated before being exported
    export_raw = list()
    for file in from_files:
        dataset = laspy.read(file)
        filtered_data = dataset[np.isin(dataset.classification, classification_filter)]
        contained_in_area = shapely.contains_xy(
            masked_area, filtered_data.x, filtered_data.y
        )
        raw_points_in_area = filtered_data[contained_in_area].points.array
        export_raw.append(raw_points_in_area)
    export_array = np.hstack(export_raw)

    export_data = laspy.LasData(export_header)
    export_data.points = laspy.ScaleAwarePointRecord(
        export_array,
        export_header.point_format,
        export_header.scales,
        export_header.offsets,
    )
    export_data.update_header()

    export_data.write(to_file)
